
/* 
 * FILE:
 * dui-initdb.h
 *
 * FUNCTION:
 * This module allows DWI to access a variety of data sources using a uniform
 * programming interface: essentially, this is a database API wrapper.  This
 * wrapper is very DWI-specific, and is not (really) meant for use by other 
 * applications (although it has been).  The API borrows conceptual elements
 * from the Postgres libpg API and the ODBC API, and so should be "inutitively
 * obvious" to experienced database programmers.   This wrapper allows a variety
 * of different database-access drivers to be operated behind it. 
 * 
 * The idea of having a 'wrapper' for database access is not new, and so its 
 * worth discussing why 'yet another' wrapper has been created.  There are
 * several reasons:
 * 
 * 1) This wrapper originially wrapped ODBC.  I didn't know ODBC very well
 *    when I did this, and I was worried that ODBC sucked.  The wrapper 
 *    allows suckage to be hidden or fixed.  
 * 2) The wrapper would allows native Postgress interfaces to be quickly
 *    developed, in case the suckage of 1 was too great.
 * 3) The wrapper allows drivers to be added for non-traditional data sources.
 *    This would allow DWI to treat these non-traditional data sources in a
 *    a database-like way: an important feature, since, to DWI, everything
 *    looks like a database.  Note: to do this properly, we would need to 
 *    move the "builder.h" and "builder.c" files from the other directory
 *    to here: the builder would be used only by SQL data sources; other
 *    data sources would have to handle key-value pairs differently.
 *    
 * This wrapper currently supports only one "provider": ODBC.
 *
 * HISTORY:
 * Created by Linas Vepstas  March 2002
 */


#ifndef DUI_INITDB_H
#define DUI_INITDB_H


typedef struct DuiDBConnection_s DuiDBConnection;
typedef struct DuiDBRecordSet_s  DuiDBRecordSet;


/* The dui_connection_new() routine establishes a connection to
 *    the indicated database, using the indicated login.  The
 *    "provider" field is the name of the driver to use. 
 *    Currently, only "odbc" and "gobject" are supported.
 *
 *    The ODBC driver expects "dbname" to specify an existing
 *    database, and "username" to specify a valid username for
 *    logging into the database.  The "authentication token"
 *    must be a valid password or other ticket that authorizes
 *    access.
 */
DuiDBConnection * dui_connection_new (const char * provider,
                                      const char * dbname, 
                                      const char * username,
                                      const char * authentication_token);

void dui_connection_free (DuiDBConnection *conn);

/* Issue a query on this connection.  FYI -- the query is actually performed
 * asynchronously; the record set may not yet contain valid data when this 
 * routine returns.  Valid data is not available until the (possibly blocking)
 * routine dui_recordset_fetch_row() is called. */
DuiDBRecordSet * dui_connection_exec (DuiDBConnection *, const char * buff);

/* Return a list of tables on this connection. */
DuiDBRecordSet * dui_connection_tables (DuiDBConnection *);

/* Get a list of columns in the named table */
DuiDBRecordSet * dui_connection_table_columns (DuiDBConnection *,
                      const char * table_name);


/* Release the record set when one is done with it */
void dui_recordset_free (DuiDBRecordSet *);

/* Prep a row.  Basically, this gets the 'next' row out of the database.
 * subsequent field fetches are for this row. Returns zero if there are 
 * no more rows.
 */
int dui_recordset_fetch_row (DuiDBRecordSet *rs);

/* Get the field value by name.  This is a little different/easier than 
 * the usual trick of getting it by column number.
 */
const char * dui_recordset_get_value (DuiDBRecordSet *, const char * fieldname);


#endif /* DUI_INITDB_H */
 
