/*
 * FILE:
 * action.h
 *
 * FUNCTION:
 * API for handling user interface activities.  "Actions" typically
 * get run as a result of the GUI user clicking on a button or performing
 * some other user interface activity.  An action stores state information
 * about what should happen when the button-click/signal/etc. occurs.
 * 
 * More precisely, actions are triggered by GTK signals issued from GTK
 * widgets.  The action stores the template of an SQL query that will be
 * run when the signal is caught.  Along with this, it stores the widgets
 * that will be queried to fill in the blanks in the SQL statement.
 * 
 * The action also stores a list of reports that need to be refreshed 
 * when the SQL query is completed.
 *
 * Not all actions need to perform SQL queries.  Some (wtokey) cause
 * data to be copied from the user interface widgets to the key-value
 * pair tree.  Others (wtoobj) copy the data to GLib-2.0 GObjects.
 *
 * HISTORY:
 * Linas Vepstas March 2002
 */

#ifndef DUI_ACTION_H
#define DUI_ACTION_H

#include <glib.h>
#include <gtk/gtkobject.h>

#include "window.h"
#include "database.h"

DuiAction * dui_action_new (DuiWindow *parent, const char *name);
void dui_action_destroy (DuiAction *);

const char * dui_action_get_name (DuiAction *);

/* Set the database connection that will be used to perform queries on */
void dui_action_set_database (DuiAction * act, DuiDatabase *db);
DuiDatabase * dui_action_get_database (DuiAction * act);


/* Specify a signal that will invoke this action. 
 * The widgetname must specify a GtkWidget on which
 * there exists a signal of name 'signalname'.
 * The widgetname must specify the name of a widget 
 * that appears in the glade file for the window that 
 * this action is a part of.
 */

void dui_action_add_signal (DuiAction *act, const char *widgettname, const char *signalname);

/* specify the sql table that will be queried in this action.
 * querytype must be one of 'select', 'update', 'insert', 'delete', 
 * 'tables' or 'columns'.
 * 'key_to-tables' must specify a KVP key that contains a table name.
 */
void dui_action_set_table (DuiAction *, const char * tables, 
                                        const char * key_to_tables,
                                        const char * querytype);

void dui_action_add_select_term (DuiAction *, 
                                const char *field_name);

void dui_action_add_update_term (DuiAction *, 
                                const char *widgetname,
                                int column,
                                const char *datakey,
                                const char *key,
                                const char *value,
                                const char *filter_name,
                                const char *field_name);

void dui_action_add_where_term (DuiAction *, 
                                const char *widgetname,
                                int column,
                                const char *datakey,
                                const char *key,
                                const char *value,
                                const char *filter_name,
                                const char *field_name,
                                const char *compare_op);

void dui_action_add_kvp_term (DuiAction *, 
                                const char *widgetname,
                                int column,
                                const char *datakey,
                                const char *source_key,
                                const char *value,
                                const char *filter_name,
                                const char *dest_key);

void dui_action_add_object_term (DuiAction *, 
                                const char *widgetname,
                                int column,
                                const char *datakey,
                                const char *key,
                                const char *value,
                                const char *filter_name,
                                const char *object_instance_name,
                                const char *object_property);


/* cross-reference to the windows that will show the results.
 * "row_name" is used to target multiple results to one report,
 * which is (for example) typically a columned tree. (GtkCTree)
 * */
void dui_action_set_report (DuiAction *, const char * name, 
                            const char * row_name);

/* Add a window that needs to be refreshed when this action is performed */
void dui_action_add_refresh (DuiAction *, const char * name);

/* Rerun the last query (and display it in a report).
 * this routine is used to refresh a window.
 */
void dui_action_rerun_last_query (DuiAction *act);


/* Add a chained action */
void dui_action_add_chain (DuiAction *act, const char * actionname);

void dui_action_do_realize (DuiAction *);
int dui_action_is_realized (DuiAction *);

/* Force the db to connect. This will cause any connect
 * errors to be reported.
 */
void dui_action_db_connect (DuiAction *act);
		  

/* Cause the indicated action to be run. */
int dui_action_run (DuiAction *act);
		  

#endif /* DUI_ACTION_H */

/* ========================== END OF FILE ============================ */
