/*
 * FILE:
 * readfile.c
 *
 * FUNCTION:
 * Parse an XML file that specifies the DWI glade-to-database coupling
 *
 * HISTORY:
 * Linas Vepstas March 2002
 */
#include "config.h"

#include <string.h>
#include <stdlib.h>

#include <glib.h>

#include <parser.h>
#include <gnome.h>

#include "perr.h"
#include "dui-initdb.h"

#include "action.h"
#include "database.h"
#include "filter.h"
#include "readfile.h"
#include "report.h"
#include "signal.h"
#include "util.h"
#include "window.h"


typedef struct DuiParser_s DuiParser;

struct DuiParser_s 
{
	/* XML file parser */
	xmlSAXHandler sax;

	DuiInterface *dui;
	
	/* Current database connection that is being used */
	DuiDatabase *database;
	
	/* Current glade window that is being parsed */
	DuiWindow *window;

	/* Current action that is being parsed */
	DuiAction  *action;

	/* Current report that is being parsed */
	DuiReport *report;

	/* Current filter definition */
	DuiFilter *filter;
};


/* =================================================================== */
/* a really cheap and dirty thing .. */

static const char * 
de_escape (const char *val)
{
	if (!val) return NULL;

	if (!strcmp (val, "&gt;")) return ">";
	if (!strcmp (val, "&lt;")) return "<";

	return val;
}

/* =================================================================== */

#define CASE_STRDUP(STR,MEMB)                        \
   if (!strcmp (attrs[i], STR)) {                    \
      MEMB = g_strdup(attrs[i+1]);                   \
   } else

#define CASE(STR,MEMB)                               \
   if (!strcmp (attrs[i], STR)) {                    \
      MEMB = attrs[i+1];                             \
   } else

#define CASE_UNKNOWN                                 \
	{                                                 \
		SYNTAX ("unknown %s attribute %s", eltname, attrs[i]); \
	}   

#define GENERIC_CHECKS                               \
	if (!attrs)                                       \
	{                                                 \
		SYNTAX ("%s must have attributes!", eltname);  \
		return;                                        \
	}

#define CHECK_NEST(fld,prnt)                         \
	if (!par->fld)                                    \
	{                                                 \
		SYNTAX ("%s can only occur within a " prnt "!", eltname);\
		return;                                        \
	}

#define REQUIRE(var,what)                            \
	if (NULL == var)                                  \
	{                                                 \
		SYNTAX ("%s must specify " what, eltname);     \
		return;                                        \
	}

#define REQUIRE2(vara,varb,what)                     \
	if ((NULL == vara) && (NULL == varb))             \
	{                                                 \
		SYNTAX ("%s must specify " what, eltname);     \
		return;                                        \
	}

#define REQUIRE3(vara,varb,varc,what)                \
	if ((NULL == vara) && (NULL == varb) && (NULL == varc))   \
	{                                                 \
		SYNTAX ("%s must specify " what, eltname);     \
		return;                                        \
	}

#define REQUIRE4(vara,varb,varc,vard,what)           \
	if ((NULL==vara) && (NULL==varb) && (NULL==varc) && (NULL==vard))   \
	{                                                 \
		SYNTAX ("%s must specify " what, eltname);     \
		return;                                        \
	}


/* =================================================================== */

static void 
handle_report (DuiParser *par, const xmlChar **attrs)
{
	DuiReport *rpt;
	static const char * eltname = "<report>";
	const char * name = NULL;
	int i;

	if (!par->window) 
	{
		SYNTAX ("must specify a glade file first!");
		return;
	}
	CHECK_NEST (window, "<window>");

	if (attrs)
	{
		i = 0;
		while (attrs[i])
		{
			CASE ("name",  name)
			CASE_UNKNOWN ;
			i += 2;
		}
	}

	/* ------------------------------------------------------------------ */
	/* Set up the widget so that when its clicked or whatever, 
	 * we run our sql statement.
	 */
	rpt = dui_report_new (par->window, name);
	par->report = rpt;
}

static void 
handle_report_end (DuiParser *par)
{
	if (!par->report) return;

	par->report = NULL;
}

/* =================================================================== */

static void 
handle_row (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<row>";
	int i;
	const char * name = NULL;
	const char * widgetname = NULL;
	const char * nest = NULL;
	const char * match_col = NULL;
	const char * match_field = NULL;
	const char * match_key = NULL;
	const char * match_value = NULL;
	int nest_level = 0;
	int mcol = -1;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");
	CHECK_NEST (report, "<report>");

	i = 0;
	while (attrs[i])
	{
		CASE ("name",         name)
		CASE ("widget",       widgetname)
		CASE ("nest",         nest)
		CASE ("match_column", match_col)
		CASE ("match_field",  match_field)
		CASE ("match_key",    match_key)
		CASE ("match_value",  match_value)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (widgetname, "a widget!");

	if (nest) nest_level = atoi (nest);
	if (match_col) mcol = atoi (match_col);
	
	dui_report_add_row (par->report, name, widgetname, nest_level, 
	                    mcol, match_field, match_key, match_value);

}

/* =================================================================== */

static void 
handle_entry (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<entry>";
	int i;
	const char * name = NULL;
	const char * widgetname = NULL;
	const char * fieldname = NULL;
	const char * column = NULL;
	const char * datakey = NULL;
	const char * arg = NULL;
	const char * key = NULL;
	const char * value = NULL;
	const char * filtername = NULL;
	int col = -1;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");
	CHECK_NEST (report, "<report>");

	i = 0;
	while (attrs[i])
	{
		CASE ("name",   name)
		CASE ("widget", widgetname)
		CASE ("column", column)
		CASE ("data",   datakey)
		CASE ("arg",    arg)
		CASE ("key",    key)
		CASE ("value",  value)
		CASE ("filter", filtername)
		CASE ("field",  fieldname)
		CASE_UNKNOWN ;
		i += 2;
	}

	// hack alert -- fixme -- 
	// a widget is not required if we are nested inside a row;
	// otherwise a widget is mandatory
	// REQUIRE (widgetname, "a widget!");
	// 
	REQUIRE3 (fieldname, value, key, "a field or a key or a value!");

	if (column) col = atoi (column);

	dui_report_add_column (par->report, widgetname, col, datakey, arg,
	                       fieldname, key, value, filtername);
}


/* =================================================================== */
/* ********************* form handlers below ************************* */
/* =================================================================== */

/* =================================================================== */

static void 
handle_form (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<form>";
	int i;
	const char * formname = NULL;
	const char * reportname = NULL;
	const char * sql_querytype = NULL;
	const char * sql_tables = NULL;
	const char * key_to_sql_tables = NULL;
	const char * row_name = NULL;
	DuiAction *act;

	if (attrs)
	{
		i = 0;
		while (attrs[i])
		{
			CASE ("name",       formname)
			CASE ("type",       sql_querytype)
			CASE ("tables",     sql_tables)
			CASE ("tables_key", key_to_sql_tables)
			CASE ("report",     reportname)
			CASE ("row",        row_name)
			CASE_UNKNOWN ;
			i += 2;
		}
	}

	/* ------------------------------------------------------------------ */
	/* Set up the widget so that when its clicked or whatever, 
	 * we run our sql statement.
	 */
	act = dui_action_new (par->window, formname);
	par->action = act;

	dui_action_set_table (act, sql_tables, key_to_sql_tables, sql_querytype);
	dui_action_set_report (act, reportname, row_name);
	dui_action_set_database (act, par->database);
}

static void 
handle_form_end (DuiParser *par)
{
	if (!par->action) return;

	par->action = NULL;
}

/* =================================================================== */

static void 
handle_submit (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<submit>";
	int i;
	const char * widgetname = NULL;
	const char * signalname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("signal", signalname)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (widgetname, "a widget!");
	REQUIRE (signalname, "a signal!");

	dui_action_add_signal (par->action, widgetname, signalname);
}

/* =================================================================== */

static void 
handle_refresh (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<refresh>";
	int i;
	const char * windowname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("window", windowname)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (windowname, "a window name!");
	dui_action_add_refresh (par->action, windowname);
}

/* =================================================================== */

static void 
handle_select (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<select>";
	int i;
	const char * fieldname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("field", fieldname)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (fieldname, "a data field!");
	dui_action_add_select_term (par->action, fieldname);
}

/* =================================================================== */

static void 
handle_insert (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<insert>";
	int i;
	const char * widgetname = NULL;
	const char * column = NULL;
	const char * datakey = NULL;
	const char * key = NULL;
	const char * value = NULL;
	const char * filtername = NULL;
	const char * fieldname = NULL;
	int colnum = -1;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("column", column)
		CASE ("data",   datakey)
		CASE ("key",    key)
		CASE ("value",  value)
		CASE ("filter", filtername)
		CASE ("field",  fieldname)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE3 (widgetname, key, value, "a widget or a key or a value!");
	REQUIRE (fieldname, "a data field!");
	
	if (column) colnum = atoi (column);

	/* set up the action so that we fetch values from it */
	dui_action_add_update_term(par->action, widgetname, 
                   colnum, datakey, key, value, filtername, fieldname);
}

/* =================================================================== */

static void 
handle_where (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<where>";
	int i;
	const char * widgetname = NULL;
	const char * column = NULL;
	const char * datakey = NULL;
	const char * fieldname = NULL;
	const char * compareop = NULL;
	const char * key = NULL;
	const char * value = NULL;
	const char * filtername = NULL;
	int colnum = -1;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("column", column)
		CASE ("data",   datakey)
		CASE ("key",    key)
		CASE ("value",  value)
		CASE ("filter", filtername)
		CASE ("field",  fieldname)
		CASE ("op",     compareop)
		CASE_UNKNOWN ;
		i += 2;
	}

	/* require either value or a key or a widget ... */
	REQUIRE3 (widgetname, key, value, "a widget or a key or a value!");
	REQUIRE (fieldname, "a data field!");

	if (column) colnum = atoi (column);

	/* set up the action so that we fetch values from it */
	compareop = de_escape (compareop);
	dui_action_add_where_term(par->action, widgetname, colnum, 
		         datakey, key, value, filtername, fieldname, compareop);

}

/* =================================================================== */

static void 
handle_chain (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<chain>";
	int i;
	const char * actionname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("form", actionname)
		CASE_UNKNOWN ;
		i += 2;
	}

	/* require an action name ... */
	REQUIRE (actionname, "a form!");

	/* set up the action so that check values */
	dui_action_add_chain (par->action, actionname);
}

/* =================================================================== */

static void 
handle_wtokey (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<wtokey>";
	int i;
	const char * widgetname = NULL;
	const char * column = NULL;
	const char * datakey = NULL;
	const char * src_key = NULL;
	const char * value = NULL;
	const char * filtername = NULL;
	const char * dst_key = NULL;
	int colnum = -1;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget",  widgetname)
		CASE ("column",  column)
		CASE ("data",    datakey)
		// CASE ("key",  src_key)
		CASE ("value",   value)
		CASE ("filter",  filtername)
		CASE ("dst_key", dst_key)
		CASE_UNKNOWN ;
		i += 2;
	}

	/* require either value or widget ... */
	REQUIRE2 (widgetname, value, "a widget or a value!");
	REQUIRE (dst_key, "a key!");

	if (column) colnum = atoi (column);

	/* set up the action so that we fetch values from it */
	dui_action_add_kvp_term(par->action, widgetname, colnum, 
	                         datakey, src_key, value, filtername, dst_key);

}

/* =================================================================== */

static void 
handle_wtoobj (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<wtoobj>";
	int i;
	const char * widgetname = NULL;
	const char * column = NULL;
	const char * datakey = NULL;
	const char * key = NULL;
	const char * value = NULL;
	const char * filtername = NULL;
	const char * objname = NULL;
	const char * propname = NULL;
	int colnum = -1;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("column", column)
		CASE ("data",   datakey)
		CASE ("key",    key)
		CASE ("value",  value)
		CASE ("filter", filtername)
		
		CASE ("object",   objname)
		CASE ("property", propname)
		CASE_UNKNOWN ;
		i += 2;
	}

	/* require either value or widget ... */
	REQUIRE3 (widgetname, key, value, "a widget or a key or a value!");
	REQUIRE (objname, "an object!");
	REQUIRE (propname, "a property!");

	if (column) colnum = atoi (column);

	/* set up the action so that we fetch values from it */
	dui_action_add_object_term(par->action, widgetname, colnum, 
	                           datakey, key, value, filtername, 
	                           objname, propname);

}

/* =================================================================== */

static void 
handle_sigaction (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<sigaction>";
	int i;
	const char * widgetname = NULL;
	const char * signalname = NULL;
	const char * sigaction = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("signal", signalname)
		CASE ("action", sigaction)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (widgetname, "a widget!");
	REQUIRE (signalname, "a valid signal name on the widget!");
	REQUIRE (sigaction,  "an action to perform!");

	dui_signal_new (par->window, widgetname, signalname, sigaction);
}

/* =================================================================== */

static void 
handle_window (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<window>";
	int i;
	const char * provider = NULL;
	const char * winname = NULL;
	const char * filepath = NULL;
	const char * root_widget = NULL;
	const char * mainattr = NULL;
	int is_main = 0;

	GENERIC_CHECKS ;

	i = 0;
	while (attrs[i])
	{
		CASE ("provider",    provider)
		CASE ("name",        winname)
		CASE ("filename",    filepath)
		CASE ("root_widget", root_widget)
		CASE ("main",        mainattr)
		CASE_UNKNOWN ;
		i += 2;
	}

	if (!winname)
	{
		REQUIRE (filepath, "a file path!");
		REQUIRE (root_widget, "a root widget!");
	}

	/* if there's already a window by this name, append to it;
	 * else create a new window */
	par->window = dui_interface_find_window_by_name (par->dui, winname);
	if (NULL == par->window)
	{
		REQUIRE (filepath, "a file path!");
		REQUIRE (root_widget, "a root widget!");

		if (mainattr && !strcasecmp (mainattr, "yes")) is_main = 1;
	
		par->window = dui_window_new (winname, par->dui);
		dui_window_set_glade (par->window, filepath, root_widget, is_main);
	}
}

static void 
handle_window_end (DuiParser *par)
{
	par->window = NULL;
}

/* =================================================================== */

static void 
handle_database (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<database>";
	int i;
	const char * name = NULL;
	const char * provider = NULL;
	const char * dbname = NULL;
	const char * hostname = NULL;
	const char * username = NULL;
	const char * passwd = NULL;

	GENERIC_CHECKS ;

	i = 0;
	while (attrs[i])
	{
		CASE_STRDUP ("name",     name)
		CASE_STRDUP ("provider", provider)
		CASE_STRDUP ("dbname",   dbname)
		CASE_STRDUP ("host",     hostname)
		CASE_STRDUP ("username", username)
		CASE_STRDUP ("passwd",   passwd)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE  (provider, "a database provider!");
	REQUIRE2 (name, dbname, "an object instance name or a database name!");
	/* NULL dbnames, usernames, etc are allowed: these might be
	 * filled in via the object interface */

	par->database = dui_database_new (name, provider, dbname, hostname, username, passwd);
	dui_interface_add_database (par->dui, par->database);
}

/* =================================================================== */

static void 
handle_lookup (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<lookup>";
	const char * key = NULL; 
	const char * val = NULL; 
	int i;

	GENERIC_CHECKS ;
	CHECK_NEST (filter, "<filter>");

	i = 0;
	while (attrs[i])
	{
		CASE_STRDUP ("key",    key)
		CASE_STRDUP ("val",    val)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (key, "a lookup key!");
	REQUIRE (val, "a value that is looked up!");
	dui_filter_add_lookup (par->filter, key, val);
}

/* =================================================================== */

static void 
handle_filter (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<filter>";
	const char * filtername = NULL; 
	const char * reversed_name = NULL; 
	int i;

	GENERIC_CHECKS ;

	i = 0;
	while (attrs[i])
	{
		CASE_STRDUP ("name",    filtername)
		CASE_STRDUP ("revname", reversed_name)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (filtername, "a name for the new filter!");
	par->filter = dui_filter_new (filtername, reversed_name);
}

static void 
handle_filter_end (DuiParser *par)
{
	par->filter = NULL;
}

/* =================================================================== */

static void 
handle_dwi (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<dwi>";
	const char * app_name = NULL;
	int i;

	if (attrs)
	{
		i = 0;
		while (attrs[i])
		{
			CASE ("name", app_name)
			CASE_UNKNOWN ;
			i += 2;
		}

		dui_interfaced_set_name (par->dui, app_name);
		
	}
}

/* =================================================================== */

static void 
fatal_err (void *user_data, const char *msg, ...)
{
	/* hack laert -- handle the elipses in the warning message */
	// DuiParser *par = (DuiInterface *) user_data;
	PERR ("fatal parsing error: %s\n", msg);
}

static void 
err (void *user_data, const char *msg, ...)
{
	// DuiParser *par = (DuiInterface *) user_data;
	PERR ("parsing error: %s\n", msg);
}

static void 
warn (void *user_data, const char *msg, ...)
{
	// DuiParser *par = (DuiInterface *) user_data;
	PERR ("parse warning: %s\n", msg);
}

static void 
start_doc (void *user_data)
{
	// DuiParser *par = (DuiInterface *) user_data;
	PINFO ("document start ptr=%p", user_data);
}

static void 
end_doc (void *user_data)
{
	// DuiParser *par = (DuiInterface *) user_data;
	PINFO ("document end");
}

/* =================================================================== */

#define DISPATCH(tok,handler)            \
   if (!strcmp (name, tok)) {            \
      handler (par, attrs);              \
   } else

static void 
start_elt (void *user_data, const xmlChar *name, const xmlChar **attrs)
{
	DuiParser *par = (DuiParser *) user_data;


	DISPATCH ("dwi",       handle_dwi)
	DISPATCH ("filter",    handle_filter)
	DISPATCH ("lookup",    handle_lookup)
	DISPATCH ("database",  handle_database)
	DISPATCH ("window",    handle_window)

	DISPATCH ("report",    handle_report)
	DISPATCH ("row",       handle_row)
	DISPATCH ("entry",     handle_entry)
	
	DISPATCH ("form",      handle_form)
	DISPATCH ("chain",     handle_chain)
	DISPATCH ("select",    handle_select)
	DISPATCH ("insert",    handle_insert)
	DISPATCH ("where",     handle_where)

	DISPATCH ("submit",    handle_submit)
	DISPATCH ("refresh",   handle_refresh)
	DISPATCH ("wtokey",    handle_wtokey)
	DISPATCH ("wtoobj",    handle_wtoobj)

	DISPATCH ("sigaction", handle_sigaction)
	{
		int i;
		PERR ("unknown element %s\n", name);
		i = 0;
		while (attrs && attrs[i])
		{
			printf ("\tattr=%s\n", attrs[i]);
			i++;
		}
	}
}


static void 
end_elt (void *user_data, const xmlChar *name)
{
	DuiParser *par = (DuiParser *) user_data;
	PINFO ("</%s>", name);

	if (!strcmp (name, "form"))
	{
		handle_form_end (par);
	}
	else
	if (!strcmp (name, "report"))
	{
		handle_report_end (par);
	}
	else
	if (!strcmp (name, "window"))
	{
		handle_window_end (par);
	}
	else
	if (!strcmp (name, "filter"))
	{
		handle_filter_end (par);
	}
}


static void
cdata_cb (void *user_data, const xmlChar *value, int len)
{
#if 0
	DuiParser *par = (DuiParser *) user_data;
	char *p;
	p = g_strndup (value, len);
	printf ("cdata duude len=%d %s\n", len, p);
	g_free (p);
#endif

}


/* =================================================================== */

static DuiParser *
dui_parser_new (void)
{
	DuiParser *par;

	par = g_new (DuiParser, 1);
	
   /* current state */
   par->dui = NULL;
   par->database = NULL;
   par->window = NULL;
   par->action = NULL;
   par->report = NULL;
   par->filter = NULL;
	
	/* initialize the SAX parser */
	bzero (&par->sax, sizeof (xmlSAXHandler));
	par->sax.startDocument = start_doc;
	par->sax.endDocument = end_doc;
	par->sax.fatalError = fatal_err;
	par->sax.error = err;
	par->sax.warning = warn;
	par->sax.startElement = start_elt;
	par->sax.endElement = end_elt;
	par->sax.characters = cdata_cb;

	return par;
}

static void 
dui_parser_destroy (DuiParser *par)
{
   /* current state */
   par->dui = NULL;
   par->database = NULL;
   par->window = NULL;
   par->action = NULL;
   par->report = NULL;
   par->filter = NULL;
	
	g_free (par);
}

void 
dui_interface_parse (DuiInterface *dui, const char * filename)
{
	DuiParser *par;
	if (!dui || !filename) return;

	if (0 == g_file_exists (filename))
	{
		FATAL ("Can't find DWI interface description file \'%s\'\n", filename);
		return;
	}
	
	par = dui_parser_new ();
	par->dui = dui;
	xmlSAXUserParseFile (&par->sax, par, filename);
	dui_parser_destroy (par);
}

/* ========================== END OF FILE ============================ */
